**Graph Query Language (GRQL)** is a domain specific language designed to allow users to write transformation rules in an easy and quick manner. It is composed of a graph language and transformation commands and constraints.

The graph language allows people to express **directed property graph patterns**.

Table of contents
-----------------
* [Primitives](#primitives)
  - [Variable, String and Rule names](#var-string-names) 
  - [Features structure](#fstruct)
  - [List of default features](#default-feats)
* [Node and Edge](#node-edge)
  - [Node](#node)
  - [Edge](#edge)
* [Graph](#graph)
  - [Description](#graph-description)
  - [Names in graph](#graph-names)
  - [Ambiguities](#graph-ambi)
* [Commands](#commands)
  - [Commands on node](#node-commands)
  - [Commands on edge](#edge-commands)
  - [Commands on node and edge](#both-commands)
* [Constraints](#constraints)
  - [What is a constraint?](#what-is-constraint)
  - [How does it work?](#how-constraint-works)
  - [Constraint types](#constraint-types)
  - [Deterministic Finite Automaton](#DFA)
  - [Linguistic examples](#constraints-ling-ex)
* [Rule anatomy](#rule-anatomy)
* [Advanced topics](#advanced-topics)
  - [Negative Application Conditions](#nac)
  - [Precedence](#precedence)


<a name='primitives'></a>
Primitives
----------

<a name='var-string-names'></a>
#### Variable, String and Rule names

**Variable names** start with an alphabetic character or an underscore (_) followed by alphanumerical characters or underscores. No other characters are allowed.

##### Examples
<pre>
_test
a_test
ate1_te3_
</pre>

**Strings** are the same as in Javascript or Python, you can use simple or double
quotes and they support escaping special characters, such as new lines (\n), tabulations (\t), ...

##### Examples
```javascript
"test"
"était"
'\tmy tabulation'
```

**Rule names** are the same as variable names but with hyphen (-) too

##### Examples
<pre>
_subject-test_rules
subject1_test-rules-2
</pre>

<a name='fstruct'></a>
#### Features structure

A features structure (FStruct) is a attribute-value matrix, which allows us to express what kind of information is stored in a graph vertex or edge.

In **GRQL**, a FStruct is expressed by a list of keys and values surrounded by by square brackets **[**.

The *keys* are basic variable names, the *values* are strings. A key can have multiple values separated by a pipe **|**. 

Each **key:values** association are separated by commas **,**.

##### Examples
<pre>
[token : "is"]
[token : "is", lemma:"be"]
[token : "is"|"was", lemma:"be"]
</pre>

As **GRQL** is a language to match small graphs in larger graphs and make transformations based on the matching, you can have more complex FStructs. In fact, you can use **!** in front of any key to tell the program you don't want to find a node or an edge with that attribute.

##### Examples
<pre>
[!token : "is"]
[cat: "verb", !lemma: "be"]
</pre>

These two examples expressed the fact that you are looking for a node (or an edge) that does not hold a **token** key with the value **is** (first example). In the second example, you are looking for a node (or an edge) with **cat** equals to *verb*, but with **lemma** different from *be*.

This allows us to write complex graph patterns.

Finally, when you do not want to specify the FStruct of a node or an edge, you just write **[]**.

<a name='default-feats'></a>
#### List of default features

Apart from the morphological field, where features are generally named. **OGRE** defines a number of predefined feature names:

- token: The token value.
- lemma: The lemma value.
- hypercat: The coarse part-of-speech.
- cat: The fine-grained part-of-speech.
- label: The dependency label.

<a name='node-edge'></a>
Node and Edge
----------

What would be a graph language without nodes and edges?

<a name='node'></a>
#### Node

A node is a conjunction of a variable and a FStruct.

1. `x` describes a node called `x`. It can have any attributes
2. `x[token:"is"]` describes a node called `x` with a token key equals to *is*.
3. `[token:"is"]` describes a node only by its FStruct. It's an anonymous node.
4. `[]` describes a general and anonymous node.

<a name='edge'></a>
#### Edge

An edge has the same syntax as a node with a direction.

1. `-[]->` expresses a general and anonymous edge.
2. `<-[]-` same as 1. but the direction is different.
3. `-x->` expresses a general edge called `x`.
4. `-x[label:"subj"]->` express a edge called `x` with a key **label** equals to *subj*.

<a name='graph'></a>
Graph
--------

<a name='graph-description'></a>
#### Description

Now that we have all the necessary tools, we can express graphs.

##### Example 1

Below is a simple graph made of two nodes and one edge. The first node is named
`x`, the second one is anonymous. The edge goes from `x` to the anonymous node.

<pre>
x -[]-> []
</pre>

##### Example 2

Below, we express a graph with one node and two out edges.

<pre>
x ( 
    -[]-> y,
    -[]-> z
)
</pre>

##### Example 3

Now, let's dive into a more complex graph:

<pre>
x[lemma: "be"] ( 
    -edge1[label:"subj"]-> y,
    -edge2[label:"obj"]-> [] <-edge3- z
)
</pre>

<div class='depgraph'>
<pre>
{
    "graph":{
        "words":[
            {"id":0,"#position":0,"label":"x"},
            {"id":1,"#position":1,"label": "y"},
            {"id":2,"#position":2,"label":"z"},
            {"id":3,"#position":3,"label":"_"}
        ],
        "links":[
            {"id":0,"source": 0,"target": 1,"label": "subj"},
            {"id":1,"source":0,"target":3,"label":"obj"},
            {"id":2, "source":2,"target":3,"label":""}
        ]
    }
}
</pre>
</div>

We have a node named `x` which should carry a attribute **lemma** with value *be*. `x` has two out edges (**edge1** and **edge2**). Moreover, the target of **edge2** is also the target of **edge3** which as a node called `z` as its source.

<a name='graph-names'></a>
#### Names in graph

1. Nodes and edges can't have same variable names in the same graph. Here `a` refers to a node and an edge. It's illegal.
<pre>
a -a-> b
</pre>

2. You can refer to the same node (or edge) by using its name as long as it's defined. Here, you make a reflexive edge on the same node `a`.
<pre>
a -[]-> a
</pre>

<a name='graph-ambi'></a>
#### Ambiguities

When dealing with multiple edges from one node, you can run into some ambiguities.

Let's imagine the graph below:
<pre>
a ( -> b -> c, -> d ) 
</pre>

Here there are two interprations:

1. `a` has two out edges going to `b` and to `d`
<pre>
a ( -> b (-> c), -> d )
</pre>

2. `a` has one out edge going to `b` and `b` has two out edges going to `c` and `d`
<pre>
a -> b ( -> c, -> d )
</pre>

Be very careful with such a thing, because you can end up with a lot of confusion. Whenever it's possible, we advice to use parenthesis (as in example 1. and 2.) or, if you want the first interpretation, to make the pattern like that:

<pre>
a ( -> d, -> b -> c )
</pre>


<a name='commands'></a>
Commands
------------

<a name='node-commands'></a>
#### Commands on node
- `remove_node(x)` remove the node called `x`.

<a name='edge-commands'></a>
#### Commands on edge
- `add_edge(a, b, fstruct)` add an edge between `a` and `b` with FStruct `fstruct`. The edge is directed from `a` to `b` ( `a -[]-> b`).
- `remove_edge(x)` remove the edge called `x`.
- `move_edge(x, a, b)` | `move_edge(x, a, b, fstruct)` move the edge called `x` and place it from `a` to `b` ( `a -x-> b` ). The function takes an optional parameter which allows the user to change (or add) properties on edge `x` after the move.
- `clone_edge(x, a, b)` | `clone_edge(x, a, b, fstruct)` same as `move_edge` but duplicate edge `x`
<!--- - `edge_status(x, boolean, boolean)` ... -->

<a name='both-commands'></a>
#### Commands on node and edge
- `add_feature(x, "key", "value")` add a property on `x` (could be a node or an edge) with name `key` and value `value`. You could add only one value per key. The function allows the use of **property copy** by using a variable name, you can access property of any element in a graph just like you do with **Python** dictionary. Let's imagine you have a graph with two nodes `a` and `b` and you want to copy the value of property `lemma` from `a` to `b`. You can do that:
<pre>
add_feature(b, "lemma", b["lemma"])
</pre>
`b["lemma"]` lets you access the value for property `lemma` in `b`.
- `remove_feature(x, "key")` remove a feature of `x` by using its key name.
<!--- - `append_feature` -->
<!--- - `modify_feature` -->


<a name='constraints'></a>
Constraints
-----------

<a name='what-is-constraint'></a>
#### What is a constraint?

<a name='how-constraint-works'></a>
#### How does it work?

<a name='constraint-types'></a>
#### Constraint types

<a name='DFA'></a>
#### Deterministic Finite Automaton

<a name='constraints-ling-ex'></a>
#### Linguistic examples


<a name='rule-anatomy'></a>
Rule anatomy
------------

Here we discuss the rule skeleton. We've already seen how to express a graph, the available commands to make transformations and the more powerful so-called constraints. So now we have everything we need to make proper rules.

#### Structures

A rule in a set of different structures:

- A *match* structure (**required**)
- A *commands* structure (**optional**)
- A *constraints* structure (**optional**)
- A *precedence* structure (**optional** and discussed in [advanced topics](#advanced-topics))
- A *negative* structure (**optional** and discussed in [advanced topics](#advanced-topics))

As you can see, the only required structure is the match. In fact, OGRE allows users to make graph matching only without any rewriting steps.

Below, we reproduce the rule skeleton in **GRQL**:

```c
rule name {
    match { }
    
    commands { }
    
    constraints { }
    
    precedence { }
    
    negative { }
}
```

Order is irrelevant. **But** you can have an infinite number of match structures. In that case, each match needs to follow the other. You **cannot** have a rule like this:

##### Incorrect rule
```c
rule name {
    match { }
    
    commands { }
    
    match { }
}
```

But you **can** have a rule like this:

##### Correct rule
```c
rule name{
    commands { }
    
    match { }
    
    match { }
}
``` 
```c
rule name{
    match { }
    
    match { }
    
    commands { }
}
```

#### Structure type

- A **match** structure holds a graph
```c
match{
    x -[]-> [token:'test']
}
```

- A **commands** structure holds commands. One command per line.
```c
commands{
    add_edge(a, b, [label:'subj'])
    remove_node(x)
}
```

- A **constraints** structure holds constraints. One constraint per line
```c
constraints{
    share_down(x, 'subj')
}
```

#### Real life example

````
#Rule for  "N pour N de Vinf".
#Ex: "l'impossibilite pour X de faire quelque chose"
#Add a subject edge between X & faire (because X is the subject of the #infinitival verb)

rule nom_a_controle_pour{
     match{
        n[hypercat:"N", lemma:"difficulté"|"impossibilité"|"possibilité"|"interdiction"|"moyen"|"obligation"|"occasion"|"opportunité"|"perspective"|"risque"|"temps"|"urgence"] (
	   -[]-> pour[lemma:"pour"] (-[label:"obj.p"]-> deepsuj[hypercat:"N"|"PRO"]),
	   -[]-> de[lemma:"de"] (-[label:"obj.p"]-> vinf[dm:"inf"])
	   )
     }

     commands{
        add_edge( vinf, deepsuj, [label:"suj"] )
     }
}
````
The subgraph we would like to match is described below:
<div class='depgraph'>
<pre>
{
   "graph":{
      "words":[
         {
            "id":0,
            "#position":0,
            "label":"n"
         },
         {
            "id":1,
            "#position":1,
            "label":"pour"
         },
         {
            "id":2,
            "#position":2,
            "label":"deepsuj"
         },
         {
            "id":3,
            "#position":3,
            "label":"de"
         },
         {
            "id":4,
            "#position":4,
            "label":"vinf"
        }
      ],
      "links":[
         {
            "id":0,
            "source":0,
            "target":1,
            "label":""
         },
         {
            "id":1,
            "source":0,
            "target":3,
            "label":""
         },
         {
            "id":2,
            "source":1,
            "target":2,
            "label":"obj.p"
         },
         {
            "id":3,
            "source":3,
            "target":4,
            "label":"obj.p"
         }
      ]
   }
}
</pre>
</div>

with the following properties:

- *n* should be a noun (coarse PoS = N) with lemma: **difficulté** or **impossibilité** or **possibilité**, etc.
- *pour* should have lemma **pour**.
- *deepsuj* should be a noun or a pronoun (coarse PoS = N or PRO).
- *de* should have lemma **de**.
- *vinf* should have morphological feature **dm** = **inf**.

An edge will be added between **deepsuj** and **vinf**.


<div class='depgraph'>
<pre>
{
   "graph":{
      "words":[
         {
            "id":0,
            "#position":0,
            "label":"n"
         },
         {
            "id":1,
            "#position":1,
            "label":"pour"
         },
         {
            "id":2,
            "#position":2,
            "label":"deepsuj"
         },
         {
            "id":3,
            "#position":3,
            "label":"de"
         },
         {
            "id":4,
            "#position":4,
            "label":"vinf"
        }
      ],
      "links":[
         {
            "id":0,
            "source":0,
            "target":1,
            "label":""
         },
         {
            "id":1,
            "source":0,
            "target":3,
            "label":""
         },
         {
            "id":2,
            "source":1,
            "target":2,
            "label":"obj.p"
         },
         {
            "id":3,
            "source":3,
            "target":4,
            "label":"obj.p"
         },
         {
            "id": 4,
            "source": 4,
            "target": 2,
            "label": "suj"
         }
      ]
   }
}
</pre>
</div>

<a name='advanced-topics'></a>
Advanced topics
---------------

<a name='nac'></a>
#### Negative Application Conditions

The **negative** structure avoids matching unwanted subgraphs. For instance, imagine that you want to add edges between subjects and controlled verbs, as in *John wants to eat an apple and drink a beer*, where **John** is the subject of **eat** and **drink**, with the following graph:

<div class='depgraph'>
<pre>
{
    "graph":{
        "words":[
            {
                "id":0,
                "#position":0,
                "label":"John"
            },
            {
                "id":1,
                "#position":1,
                "label":"wants"
            },
            {
                "id":2,
                "#position":2,
                "label":"eat"
            },
            {
                "id":3,
                "#position":3,
                "label":"apple"
            },
            {
                "id":4,
                "#position":4,
                "label":"and"
            },
            {
                "id":5,
                "#position":5,
                "label":"drink"
            },
            {
                "id":6,
                "#position":6,
                "label":"beer"
            }
        ],
        "links":[
            {
                "id":0,
                "source":1,
                "target":0,
                "label":"suj"
            },
            {
                "id":1,
                "source":1,
                "target":2,
                "label":"obj"
            },
            {
                "id":2,
                "source":2,
                "target":3,
                "label":"obj"
            },
            {
                "id":3,
                "source":1,
                "target":4,
                "label":"coord"
            },
            {
                "id": 4,
                "source": 4,
                "target": 5,
                "label": "dep.coord"
            },
            {
                "id": 5,
                "source": 5,
                "target": 6,
                "label": "obj"
            }
        ]
    }
}
</pre>
</div>

The simplest rule that can be designed is:

```c
rule name {
    match { 
        suj <-[label:"suj"]- v[hypercat:"V"] (
            -[label: "obj"]-> vinf[cat: "VINF"],
            -[label: "coord"]-> and[hypercat: "C"] -[label: "dep.coord"]-> vinf2[cat: "VINF"]
        )
    }
    
    commands { 
        add_edge(vinf, suj, [label: "suj"])
        add_edge(vinf2, suj, [label: "suj"])
    }
}
```

But this rule will match graphs such as:

<div class='depgraph'>
<pre>
{
    "graph":{
        "words":[
            {
                "id":0,
                "#position":0,
                "label":"John"
            },
            {
                "id":1,
                "#position":1,
                "label":"wants"
            },
            {
                "id":2,
                "#position":2,
                "label":"eat"
            },
            {
                "id":3,
                "#position":3,
                "label":"apple"
            },
            {
                "id":4,
                "#position":4,
                "label":"and"
            },
            {
                "id":5,
                "#position":5,
                "label":"Peter"
            },
            {
                "id":6,
                "#position":6,
                "label":"drink"
            },
            {
                "id":7,
                "#position":7,
                "label":"beer"
            }
        ],
        "links":[
            {
                "id":0,
                "source":1,
                "target":0,
                "label":"suj"
            },
            {
                "id":1,
                "source":1,
                "target":2,
                "label":"obj"
            },
            {
                "id":2,
                "source":2,
                "target":3,
                "label":"obj"
            },
            {
                "id":3,
                "source":1,
                "target":4,
                "label":"coord"
            },
            {
                "id": 4,
                "source": 4,
                "target": 6,
                "label": "dep.coord"
            },
            {
                "id": 4,
                "source": 6,
                "target": 5,
                "label": "suj"
            },
            {
                "id": 5,
                "source": 6,
                "target": 7,
                "label": "obj"
            }
        ]
    }
}
</pre>
</div>

because we cannot block the fact that **drink** (*vinf2*) already has a subject.

Negative application conditions come into play here. Modifying the rule like this will avoid the aforementionned situation.

```c
rule name {
    match { 
        suj <-[label:"suj"]- v[hypercat:"V"] (
            -[label: "obj"]-> vinf[cat: "VINF"],
            -[label: "coord"]-> and[hypercat: "C"] -[label: "dep.coord"]-> vinf2[cat: "VINF"]
        )
    }
    
    negative {
        vinf2 -[label: "suj"]-> []
    }
    
    commands { 
        add_edge(vinf, suj, [label: "suj"])
        add_edge(vinf2, suj, [label: "suj"])
    }
}
```

This says: *vinf2* should not have a subject (= an edge subject connected to any node).

You can add as many negative structures as you want.

<a name='precedence'></a>
#### Precedence

The **precedence** structure imposes a linear order to its nodes. Looking back at the previous example (above), if we want to force *suj* to be before *v* and *v* before *vinf*, we can add:

```c
rule name {
    match { 
        suj <-[label:"suj"]- v[hypercat:"V"] (
            -[label: "obj"]-> vinf[cat: "VINF"],
            -[label: "coord"]-> and[hypercat: "C"] -[label: "dep.coord"]-> vinf2[cat: "VINF"]
        )
    }
    
    negative {
        vinf2 -[label: "suj"]-> []
    }
    
    precedence {
        suj < v < vinf
    }
    
    commands { 
        add_edge(vinf, suj, [label: "suj"])
        add_edge(vinf2, suj, [label: "suj"])
    }
}
```
